#define _XOPEN_SOURCE
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/Container.h>
#include <Xm/IconG.h>
#include <Xm/ScrolledW.h>

#include <Xmd/Icon.h>
#include <Xmd/Exec.h>
#include <Xmd/Buffer.h>
#include <Xmd/StringMap.h>

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>

#include "icons/generic.xbm"
#include "icons/generic-small.xbm"
#include "icons/icon.xbm"
#include "icons/disk.xbm"
#include "icons/disk-small.xbm"
#include "icons/rom.xbm"
#include "icons/rom-small.xbm"
#include "icons/tty.xbm"
#include "icons/tty-small.xbm"
#include "icons/filesystem.xbm"
#include "icons/filesystem-small.xbm"

XtAppContext application;
XmdStringMap *devices;
Widget       container;
Pixmap       iconsLarge[5] = { 0 };
Pixmap       iconsSmall[5] = { 0 };

typedef enum {
	DeviceKindGeneric,
	DeviceKindDisk,
	DeviceKindRom,
	DeviceKindFilesystem,
	DeviceKindTTY
} DeviceKind;

typedef struct {
	Widget     widget;
	DeviceKind kind;
	String     name;
	String     mountPoint;
	String     label;
} Device;

void     loadIconPixmaps (void);
void     refreshDevices  (void);
Device * DeviceNew           (String name, DeviceKind kind);
void     DeviceDelete        (ConstString name);
void     DeviceFree          (Device *this);
void     DeviceSetKind       (Device *this, DeviceKind kind);
void     DeviceSetMountPoint (Device *this, ConstString mountPoint);
void     DeviceSetLabel      (Device *this, ConstString label);
int readBlockDevice (FILE *stream, Device **);
int readLsblkPair   (FILE *stream, String *key, String *value);

int main (int argc, char *argv[]) {
	devices = XmdStringMapNew();

	Widget window = XtVaAppInitialize (
		&application, "Devices",
		NULL, 0,
		&argc, argv,
		NULL,
		XmNtitle, "Devices",
		XmNiconName, "Devices",
		NULL);
	Pixmap iconPixmap = XmdLoadBitmapIcon(window, icon);
	XtVaSetValues (
		window,
		XmNiconPixmap, iconPixmap,
		NULL);
	container = XtVaCreateManagedWidget (
		"container", xmContainerWidgetClass, window,
		XmNlayoutType, XmSPATIAL,
		XmNspatialStyle, XmGRID,
		NULL);
		
	loadIconPixmaps();
	refreshDevices();
	
	XtRealizeWidget(window);
	XtAppMainLoop(application);
}

void loadIconPixmaps (void) {
	#define dev(kind, name) \
		iconsLarge[DeviceKind##kind] = XmdLoadBitmapIcon(container, name); \
		iconsSmall[DeviceKind##kind] = XmdLoadBitmapIcon(container, name##_small)
	dev(Generic, generic);
	dev(Disk, disk);
	dev(Rom, rom);
	dev(Filesystem, filesystem);
	dev(TTY, tty);
}

void refreshDevices (void) {
	XtUnmanageChild(container);
	
	pid_t child = 0;
	FILE *stream = XmdVaPipedExecPath (
		"lsblk", &child, "r",
		"", "-P", "-o", "NAME,TYPE,MOUNTPOINTS,LABEL", NULL);
	if (stream == NULL) {
		/* TODO error */
		return;
	}
	Device *device = NULL;
	while (readBlockDevice(stream, &device) != EOF) {
		/* do something with device? */
	}
	fclose(stream);
	waitpid(child, NULL, 0);
			
	XtManageChild(container);
}

int readBlockDevice (FILE *stream, Device **device) {
	DeviceKind kind   = DeviceKindGeneric;
	String name       = NULL;
	String mountPoint = NULL;
	String label      = NULL;
	
	int ch = 0;
	while (1) {
		String key   = NULL;
		String value = NULL;
		ch = readLsblkPair(stream, &key, &value);
		
		if (strcmp(key, "NAME") == 0) {
			XtFree(name);
			name = value;
		} else if (strcmp(key, "TYPE") == 0) {
			if (strcmp(value, "disk") == 0) {
				kind = DeviceKindDisk;
			} else if (strcmp(value, "rom") == 0) {
				kind = DeviceKindRom;
			} else if (strcmp(value, "part") == 0) {
				kind = DeviceKindFilesystem;
			}
			XtFree(value);
		} else if (strcmp(key, "MOUNTPOINTS") == 0) {
			XtFree(mountPoint);
			mountPoint = value;
		} else if (strcmp(key, "LABEL") == 0) {
			XtFree(label);
			label = value;
		} else {
			XtFree(value);
		}
		XtFree(key);
		
		while ((ch = fgetc(stream)) == ' ');
		if (ch == EOF || ch == '\n') break;
		ungetc(ch, stream);
	}
	
	if (name != NULL) {
		*device = XmdStringMapGet(devices, name);
		if (*device == NULL) {
			*device = DeviceNew(name, kind);
		} else {
			DeviceSetKind(*device, kind);
		}
		DeviceSetMountPoint(*device, mountPoint);
		DeviceSetLabel(*device, label);
	}
	
	XtFree(name);
	XtFree(mountPoint);
	XtFree(label);
	return ch;
}

int readLsblkPair (FILE *stream, String *key, String *value) {
	int  ch       = 0;
	int  escapeCh = 0;
	char null     = 0;
	
	*key   = NULL;
	*value = NULL;
	
	XmdBuffer *keyBuffer = XmdBufferNew(char);
	while (isalpha(ch = fgetc(stream))) XmdBufferPush(keyBuffer, &ch);
	XmdBufferPush(keyBuffer, &null);
	*key = XmdBufferBreak(keyBuffer);
	
	fgetc(stream); /* skip = */
	
	XmdBuffer *valueBuffer = XmdBufferNew(char);
	enum {
		stateGetString,
		stateGetEscape,
		
	} state = stateGetString;
	while (1) {
		ch = fgetc(stream);
		if (ch == EOF) goto end;
		
		switch (state) {
		case stateGetString:
			switch (ch) {
			case '"':  goto end;
			case '\\': state = stateGetEscape;          break;
			default:   XmdBufferPush(valueBuffer, &ch); break;
			}
			break;
			
		case stateGetEscape:
			ch = fgetc(stream); /* skip x */
			int digit = tolower(ch);
			if (
				(digit < '0' || digit > '9') &&
				(digit < 'a' || digit > 'f')
			) {
				ungetc(ch, stream);
				XmdBufferPush(valueBuffer, &escapeCh);
				state = stateGetString;
				break;
			}
			digit -= '0';
			if (digit > 9) digit -= 'a' - '9' - 1;
			escapeCh *= 16;
			escapeCh += digit;
			break;
		}
	}
	end:
	XmdBufferPush(valueBuffer, &null);
	*value = XmdBufferBreak(valueBuffer);
	if (strlen(*value) == 0) {
		XtFree(*value);
		*value = NULL;
	}
	return ch;
}

Device *DeviceNew (String name, DeviceKind kind) {
	Device *this = XtNew(Device);
	this->mountPoint = NULL;
	this->label      = NULL;
	this->name       = XtNewString(name);
	XmString string = XmStringCreateLocalized(name);
	this->widget = XtVaCreateManagedWidget (
		"device", xmIconGadgetClass, container,
		XmNalignment, XmALIGNMENT_CENTER,
		XmNlabelString, string,
		NULL);
	XmStringFree(string);
	DeviceSetKind(this, kind);
	XmdStringMapSet(devices, name, this);
	return this;
}

void DeviceDelete (ConstString name) {
	DeviceFree(XmdStringMapSet(devices, name, NULL));
}

void DeviceFree (Device *this) {
	if (this == NULL) return;
	XtDestroyWidget(this->widget);
	XtFree((char *)(this->name));
	XtFree((char *)(this->mountPoint));
	XtFree((char *)(this->label));
	XtFree((char *)(this));
}

void DeviceSetKind (Device *this, DeviceKind kind) {
	XtVaSetValues (
		this->widget,
		XmNlargeIconPixmap, iconsLarge[kind],
		XmNsmallIconPixmap, iconsSmall[kind],
		NULL);
	this->kind = kind;
}

void DeviceSetMountPoint (Device *this, ConstString mountPoint) {
	XtFree(this->mountPoint);
	this->mountPoint = XtNewString(mountPoint);
}

void DeviceSetLabel (Device *this, ConstString label) {
	XtFree(this->label);
	this->label = XtNewString(label);
	
	XmString string = NULL;
	if (label == NULL) {
		string = XmStringCreateLocalized(this->name);
	} else {
		string = XmStringCreateLocalized(this->label);
	}
	XtVaSetValues (
		this->widget,
		XmNlabelString, string,
		NULL);
	XmStringFree(string);
}
